/* -*-mode: c; indent-tabs-mode: nil; c-basic-offset: 2; -*-
 */

/**
 * See http://sourceforge.net/projects/modsetenvifplus/ for further
 * details about mod_setenvifplus.
 *
 * Copyright (C) 2010-2011 Pascal Buchbinder
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/************************************************************************
 * Version
 ***********************************************************************/
static const char revision[] = "$Id: mod_setenvifplus.c,v 1.24 2011/12/05 19:59:58 pbuchbinder Exp $";
static const char g_revision[] = "0.13";

/************************************************************************
 * Includes
 ***********************************************************************/
/* openssl */

/* apr */
#include "apr.h"
#include "apr_strings.h"
#include "apr_strmatch.h"
#include "apr_lib.h"
#include "apr_base64.h"

#define APR_WANT_STRFUNC
#include "apr_want.h"

/* apache */
//#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include <util_filter.h>

/************************************************************************
 * defines
 ***********************************************************************/
#define SP_LOG_PFX(id)  "mod_setenvifplus("#id"): "

enum sp_special {
  SP_SPECIAL_NOT,
  SP_SPECIAL_REMOTE_ADDR,
  SP_SPECIAL_REMOTE_HOST,
  SP_SPECIAL_REQUEST_URI,
  SP_SPECIAL_REQUEST_QUERY,
  SP_SPECIAL_REQUEST_USER,
  SP_SPECIAL_REQUEST_METHOD,
  SP_SPECIAL_REQUEST_PROTOCOL,
  SP_SPECIAL_SERVER_ADDR,
  SP_SPECIAL_RESPONSE_STATUS
};

enum sp_action {
  SP_ACTION_SET,
  SP_ACTION_UNSET,
  SP_ACTION_CHANGE,
  SP_ACTION_ADD
};

/************************************************************************
 * structures
 ***********************************************************************/
typedef struct {
  const char *name;             /* attribute name */
  ap_regex_t *pnamereg;         /* compiled attribute name (optional regex) */
  const char *regex;            /* regex to match against */
  ap_regex_t *preg;             /* compiled regex */
  const apr_strmatch_pattern *pattern; /* non-regex pattern to match */
  const char *variable;         /* env var name to set (or unset) */
  const char *value;            /* env var value to set */ 
  enum sp_special special_type; /* is it a "special" attribute ? */
  int icase;                    /* ignoring case? */
} sp_std_entry_t;

typedef struct {
  const char *name;
  ap_regex_t *preg;             /* compiled regex (for SP_ACTION_CHANGE only) */
  const char *value;
  const char *condition;
  enum sp_action action;
} sp_hdr_entry_t;

typedef struct {
  const char *header;
  ap_regex_t *preg;
  const char *value;
} sp_usr_entry_t;

typedef struct {
  char *path;
  apr_array_header_t *std_conditionals;
  apr_array_header_t *std_late_conditionals;
  apr_array_header_t *std_res_conditionals;
  apr_array_header_t *std_req_header;
  apr_array_header_t *std_late_req_header;
  apr_array_header_t *std_res_header;
  apr_array_header_t *user;
} sp_config_t;

/* functions to replace patterns */
typedef struct {
  char *identifier;                     /** name of the function inc. opening
                                            bracket, e.g. "e64(" */
  int len;
  char close;                           /** end marker, e.g. ')' */
  char *(*func)(request_rec *, char *); /** pointer to the function */
} sp_function_t;

typedef struct sp_url_escape_seq {
  char c;
  char *esc;
  apr_size_t len;
} sp_url_escape_seq_t;

/************************************************************************
 * globals
 ***********************************************************************/
module AP_MODULE_DECLARE_DATA setenvifplus_module;
#define SP_ICASE_MAGIC  ((void *)(&setenvifplus_module))

/************************************************************************
 * private
 ***********************************************************************/

static int sp_hex2c(const char *x) {
  int i, ch;
  ch = x[0];
  if (apr_isdigit(ch)) {
    i = ch - '0';
  }else if (apr_isupper(ch)) {
    i = ch - ('A' - 10);
  } else {
    i = ch - ('a' - 10);
  }
  i <<= 4;
  
  ch = x[1];
  if (apr_isdigit(ch)) {
    i += ch - '0';
  } else if (apr_isupper(ch)) {
    i += ch - ('A' - 10);
  } else {
    i += ch - ('a' - 10);
  }
  return i;
}

#define SP_ISHEX(x) (((x >= '0') && (x <= '9')) ||  \
                     ((x >= 'a') && (x <= 'f')) ||  \
                     ((x >= 'A') && (x <= 'F')))


/**
 * url escaping/encoding (%xx, '+')
 */
static char *sp_url_encoding(apr_pool_t *pool, const char *x) {
  int i = 0;
  int k = 0;
  int len = strlen(x);
  char *c = apr_pcalloc(pool, (3 * len) + 1);
  unsigned char *in = (unsigned char *)x;
  while(in[i]) {
    if(strchr("?\n /;%=\"'.:@\\&+", in[i]) != NULL) {
      sprintf(&c[k], "%%%02x", in[i]);
      k+=3;
    } else {
      c[k] = in[i];
      k++;
    }
    i++;
  }
  c[k] = '\0';
  return c;
}

/**
 * url unescaping/decoding (%xx, '+')
 */
static int sp_url_decoding(char *x, int *error) {
  int i, j, ch;
  if(x == 0) {
    return 0;
  }
  if(x[0] == '\0') {
    return 0;
  }
  for(i = 0, j = 0; x[i] != '\0'; i++, j++) {
    ch = x[i];
    if(ch == '%') {
      if(SP_ISHEX(x[i + 1]) && SP_ISHEX(x[i + 2])) {
        ch = sp_hex2c(&x[i + 1]);
        i += 2;
      } else {
        (*error)++;
      }
    } else if(ch == '+') {
      ch = ' ';
    }
    x[j] = ch;
  }
  x[j] = '\0';
  return j;
}

static int sp_is_header_regex(apr_pool_t *p, const char* name) {
  /* If a Header name contains characters other than:
   *    -,_,[A-Z\, [a-z] and [0-9].
   * assume the header name is a regular expression.
   */
  ap_regex_t *preg = ap_pregcomp(p, "^[-A-Za-z0-9_]*$", (AP_REG_EXTENDED|AP_REG_NOSUB));
  ap_assert(preg != NULL);
  if(ap_regexec(preg, name, 0, NULL, 0)) {
    return 1;
  }
  return 0;
}

/* If the input string does not take advantage of regular
 * expression metacharacters, return a pointer to an equivalent
 * string that can be searched using apr_strmatch().  (The
 * returned string will often be the input string.  But if
 * the input string contains escaped characters, the returned
 * string will be a copy with the escapes removed.)
 */
static const char *sp_non_regex_pattern(apr_pool_t *p, const char *s) {
  const char *src = s;
  int escapes_found = 0;
  int in_escape = 0;
  
  while (*src) {
    switch (*src) {
    case '^':
    case '.':
    case '$':
    case '|':
    case '(':
    case ')':
    case '[':
    case ']':
    case '*':
    case '+':
    case '?':
    case '{':
    case '}':
      if(!in_escape) {
        return NULL;
      }
      in_escape = 0;
      break;
    case '\\':
      if(!in_escape) {
        in_escape = 1;
        escapes_found = 1;
      } else {
        in_escape = 0;
      }
      break;
    default:
      if(in_escape) {
        return NULL;
      }
      break;
    }
    src++;
  }
  if(!escapes_found) {
    return s;
  } else {
    char *unescaped = (char *)apr_palloc(p, src - s + 1);
    char *dst = unescaped;
    src = s;
    do {
      if(*src == '\\') {
        src++;
      }
    } while((*dst++ = *src++));
    return unescaped;
  }
}

/**
 * replaces <start>var<end> (e.g. ${var}) entries by the value of the
 * variable within the vars table
 * retruns 1 (success) if all start markers could be replaced
 */
static int sp_reslove_variable(request_rec *r,
                               const char *start_marker, int start_marker_len, const char end_marker,
                               char **string,
                               char *(*func)(request_rec *, char *)) {
  int i;
  int start;
  int line_end;
  char *var_name;
  char *new_line = *string;
  char *line = *string;
  const char *val;
  
 once_again:
  i = 0;
  while(line[i] != 0) {
    if(strncmp(&line[i], start_marker, start_marker_len) == 0) {
      line_end = i;
      i+=start_marker_len;
      start = i;
      while((line[i] != 0) && (line[i] != end_marker)) {
        i++;
      }
      if(line[i] != end_marker) {
        /* no end found */
        break;
      } else {
        var_name = apr_pstrndup(r->pool, &line[start], i - start);
        val = func(r, var_name); 
        if(val) {
          line[line_end] = 0;
          i++;
          new_line = apr_pstrcat(r->pool, line, val, &line[i], NULL);
          line = new_line;
          goto once_again;
        }      
      }
    }
    i++;
  }
  if(!new_line[0] || strstr(new_line, start_marker)) {
    return 0;
  }
  *string = new_line;
  return 1;
}

/* replaces the env variable by its value */
static char *sp_func_variable(request_rec *r, char *variable) {
  return apr_pstrdup(r->pool, apr_table_get(r->subprocess_env, variable));
  
}

static char *sp_func_eURL(request_rec *r, char *string) {
  return sp_url_encoding(r->pool, string);
}

static char *sp_func_dURL(request_rec *r, char *string) {
  int error;
  char *c = apr_pstrdup(r->pool, string);
  sp_url_decoding(c, &error);
  return c;
}

static char *sp_func_e64(request_rec *r, char *string) {
  return ap_pbase64encode(r->pool, string);
}

static char *sp_func_d64(request_rec *r, char *string) {
  return ap_pbase64decode(r->pool, string);
}

static char *sp_func_RND(request_rec *r, char *num) {
  char *buf;
  apr_int64_t len = apr_atoi64(num);
  unsigned char *secret = apr_pcalloc(r->pool, len);
#if APR_HAS_RANDOM
  apr_generate_random_bytes(secret, len);
#else
  ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                SP_LOG_PFX(030)"$RND() can't generate random data."
                " APR random number support is missing!");
#endif
  buf = (char *)apr_pcalloc(r->pool, 1 + apr_base64_encode_len(len));
  apr_base64_encode(buf, (const char *)secret, len);
  buf[len] = '\0';
  return buf;
}

/* function list */
static sp_function_t sp_function_list[]= {
  { "$e64(",  5, ')', &sp_func_e64 },
  { "$d64(",  5, ')', &sp_func_d64 },
  { "$dURL(", 6, ')', &sp_func_dURL },
  { "$eURL(", 6, ')', &sp_func_eURL },
  { "$RND(",  5, ')', &sp_func_RND },
  { NULL,     0, 0,   NULL }
};

/**
 * implements the kown functions, e.g. base64 encoding
 */
static char *sp_process_functions(request_rec *r, const char *replaced) {
  sp_function_t *e = sp_function_list;
  char *string = apr_pstrdup(r->pool, replaced);
  while(e && e->identifier) {
    sp_reslove_variable(r, e->identifier, e->len, e->close, &string, e->func);
    e++;
  }
  return string;
}

/**
 * resolves the variable value and (un-)sets it
 */
static void sp_match(request_rec *r, sp_std_entry_t *b, const char *val) {
  if(val) {
    ap_regmatch_t regm[AP_MAX_REG_MATCH];
    apr_size_t val_len = strlen(val);
    if((b->pattern && apr_strmatch(b->pattern, val, val_len)) ||
       (!b->pattern && !ap_regexec(b->preg, val, AP_MAX_REG_MATCH, regm, 0))) {
      if(b->variable[0] == '!') {
        apr_table_unset(r->subprocess_env, &b->variable[1]);
      } else {
        char *replaced = NULL;
        if(!b->pattern) {
          /* 1) respolve sub regex */
          replaced = ap_pregsub(r->pool, b->value, val, AP_MAX_REG_MATCH, regm);
        } else {
          /* literal string (no $1 .. $9) */
          replaced = apr_pstrdup(r->pool, b->value);
        }
        if(replaced) {
          /* 2) sp_reslove_variable() resolves all ${xxx} variables */
          if(sp_reslove_variable(r, "${", strlen("${"), '}', &replaced, &sp_func_variable)) {
            /* 3) process functions */
            replaced = sp_process_functions(r, replaced);
            apr_table_set(r->subprocess_env, b->variable, replaced);
          }
        }
      }
    }
  }
  return;
}

/**
 * Determines the attriubte to match and calls sp_match() in order
 * to (un-)set the variable
 */
static int sp_setenv(request_rec *r, apr_table_t *headers, apr_array_header_t *conditionals) {
  int i;
  sp_std_entry_t *entries = (sp_std_entry_t *)conditionals->elts;
  
  for(i = 0; i < conditionals->nelts; ++i) {
    sp_std_entry_t *b = &entries[i];
    const char *val = NULL;
    
    switch (b->special_type) {
    case SP_SPECIAL_REMOTE_ADDR:
      val = r->connection->remote_ip;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_SERVER_ADDR:
      val = r->connection->local_ip;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REMOTE_HOST:
      val =  ap_get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME, NULL);
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_URI:
      val = r->uri;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_METHOD:
      val = r->method;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_PROTOCOL:
      val = r->protocol;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_QUERY:
      val = r->args;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_REQUEST_USER:
      val = r->user;
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_RESPONSE_STATUS:
      val = apr_psprintf(r->pool, "%d", r->status);
      sp_match(r, b, val);
      break;
    case SP_SPECIAL_NOT:
      if (b->pnamereg) {
        /* Matching headers_in against a regex. Iterate through
         * the headers until we find a match or run out of
         * headers.
         */
        int j;
        const apr_array_header_t *arr = apr_table_elts(headers);
        const apr_table_entry_t *elts = (const apr_table_entry_t *)arr->elts;
        for(j = 0; j < arr->nelts; ++j) {
          if(!ap_regexec(b->pnamereg, elts[j].key, 0, NULL, 0)) {
            val = elts[j].val;
            sp_match(r, b, val);
          }
        }
      } else {
        /* Not matching against a regex */
        val = apr_table_get(headers, b->name);
        if(val == NULL) {
          val = apr_table_get(r->subprocess_env, b->name);
        }
        sp_match(r, b, val);
      }
    }
  }  
  return DECLINED;
}

static int sp_condition(apr_table_t *subprocess_env, const char *condition) {
  if(condition[0] == '!') {
    if(!apr_table_get(subprocess_env, &condition[1])) {
      return 1;
    }
  } else {
    if(apr_table_get(subprocess_env, condition)) {
      return 1;
    }
  }
  return 0;
}

/* ChangeRequestHeaderPlus, ChangeResponseHeaderPlus */
static void sp_change_header(request_rec *r, apr_table_t *headers, apr_table_t *change_rules) {
  ap_regmatch_t regm[AP_MAX_REG_MATCH];
  apr_table_t *changed = apr_table_make(r->pool, 20); /* modified headers */
  apr_table_t *orig = apr_table_make(r->pool, 20);    /* unchanged headers */
  int i;
  apr_table_entry_t *header = (apr_table_entry_t *)apr_table_elts(headers)->elts;
  for(i = 0; i < apr_table_elts(headers)->nelts; i++) {
    int match = 0;
    int k;
    apr_table_entry_t *rule = (apr_table_entry_t *)apr_table_elts(change_rules)->elts;
    for(k = 0; k < apr_table_elts(change_rules)->nelts; k++) {
      sp_hdr_entry_t *b = (sp_hdr_entry_t *)rule[k].val;
      if(strcasecmp(b->name, header[i].key) == 0) {
        if(!ap_regexec(b->preg, header[i].val, AP_MAX_REG_MATCH, regm, 0)) {
          char *replaced = ap_pregsub(r->pool, b->value, header[i].val, AP_MAX_REG_MATCH, regm);
          sp_reslove_variable(r, "${", strlen("${"), '}', &replaced, &sp_func_variable);
          replaced = sp_process_functions(r, replaced);
          apr_table_addn(changed, header[i].key, replaced);
          match = 1;
          break;
        }
      }
    }
    if(!match) {
      apr_table_addn(orig, header[i].key, header[i].val);
    }
  }
  apr_table_clear(headers);
  header = (apr_table_entry_t *)apr_table_elts(orig)->elts;
  for(i = 0; i < apr_table_elts(orig)->nelts; i++) {
    apr_table_addn(headers, header[i].key, header[i].val);
  }
  header = (apr_table_entry_t *)apr_table_elts(changed)->elts;
  for(i = 0; i < apr_table_elts(changed)->nelts; i++) {
    apr_table_addn(headers, header[i].key, header[i].val);
  }
}


static int sp_header(request_rec *r, apr_table_t *headers, apr_array_header_t *list) {
  apr_table_t *change_rules = NULL;
  int i;
  sp_hdr_entry_t *entries = (sp_hdr_entry_t *)list->elts;
  for(i = 0; i < list->nelts; ++i) {
    char *val;
    sp_hdr_entry_t *b = &entries[i];
    switch (b->action) {
    case SP_ACTION_SET:
      val = apr_pstrdup(r->pool, b->value);
      if(sp_reslove_variable(r, "${", strlen("${"), '}', &val, &sp_func_variable)) {
        if(b->condition) {
          if(sp_condition(r->subprocess_env, b->condition)) {
            apr_table_set(headers, b->name, val);
          }
        } else {
          apr_table_set(headers, b->name, val);
        }
      }
      break;
    case SP_ACTION_UNSET:
      if(b->condition) {
        if(sp_condition(r->subprocess_env, b->condition)) {
          apr_table_unset(headers, b->name);
        }
      } else {
        apr_table_unset(headers, b->name);
      }
      break;
    case SP_ACTION_ADD:
      val = apr_pstrdup(r->pool, b->value);
      if(sp_reslove_variable(r, "${", strlen("${"), '}', &val, &sp_func_variable)) {
        if(b->condition) {
          if(sp_condition(r->subprocess_env, b->condition)) {
            apr_table_add(headers, b->name, val);
          }
        } else {
          apr_table_add(headers, b->name, val);
        }
      }
      break;
    case SP_ACTION_CHANGE:
      if(!change_rules) {
        change_rules = apr_table_make(r->pool, 1);
      }
      apr_table_addn(change_rules, apr_psprintf(r->pool, "%d", i), (char *)b);
      break;
    }
  }
  if(change_rules) {
    sp_change_header(r, headers, change_rules);
  }
  return DECLINED;
}

/************************************************************************
 * handlers
 ***********************************************************************/

static int sp_fixup(request_rec *r) {
  sp_config_t *conf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  if(conf) {
    sp_setenv(r, r->headers_in, conf->std_late_conditionals);
    sp_header(r, r->headers_in, conf->std_late_req_header);
  }
  return DECLINED;
}

static int sp_user_id(request_rec *r) {
  sp_config_t *conf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  if(conf) {
    ap_regmatch_t regm[AP_MAX_REG_MATCH];
    apr_array_header_t *c = conf->user;
    int i;
    sp_usr_entry_t *entries = (sp_usr_entry_t *)c->elts;
    for(i = 0; i < c->nelts; ++i) {
      sp_usr_entry_t *b = &entries[i];
      const char *hdr = apr_table_get(r->headers_in, b->header);
      if(hdr) {
        if(!ap_regexec(b->preg, hdr, AP_MAX_REG_MATCH, regm, 0)) {
          char *replaced = ap_pregsub(r->pool, b->value, hdr, AP_MAX_REG_MATCH, regm);
          if(replaced) {
            r->user = replaced;
            r->ap_auth_type = (char*)apr_pstrdup(r->pool, "delegated");
            return OK;
          }
        }
      }
    }
  }
  return DECLINED;
}

static int sp_header_parser(request_rec *r) {
  sp_config_t *conf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  if(conf) {
    sp_setenv(r, r->headers_in, conf->std_conditionals);
    sp_header(r, r->headers_in, conf->std_req_header);
  }
  return DECLINED;
}

static int sp_post_read_request(request_rec *r) {
  sp_config_t *conf = ap_get_module_config(r->server->module_config, &setenvifplus_module);
  if(conf) {
    sp_setenv(r, r->headers_in, conf->std_conditionals);
    sp_header(r, r->headers_in, conf->std_req_header);
  }
  return DECLINED;
}

#ifdef SP_INTERNAL_TEST
/** test handler printing all variables */
static int sp_handler(request_rec * r) {
  if(strcmp(r->handler, "setenvifplus-test")) {
    return DECLINED;
  } else {
    int i;
    apr_table_entry_t *entry = (apr_table_entry_t *)apr_table_elts(r->subprocess_env)->elts;
    ap_set_content_type(r, "text/plain");
    for(i = 0; i < apr_table_elts(r->subprocess_env)->nelts; i++) {
      ap_rprintf(r, "var: <%s><%s>\n", entry[i].key, entry[i].val);
    }
    entry = (apr_table_entry_t *)apr_table_elts(r->headers_in)->elts;
    ap_set_content_type(r, "text/plain");
    apr_table_set(r->headers_out, "X-SetEnvIfPlus",
                  apr_psprintf(r->pool, "testhandler=%s", g_revision));
    for(i = 0; i < apr_table_elts(r->headers_in)->nelts; i++) {
      ap_rprintf(r, "hdr: <%s><%s>\n", entry[i].key, entry[i].val);
    }
  }
  return OK;
}
#endif

/** finalize configuration */
static int sp_post_config(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp,
                            server_rec *bs) {
  ap_add_version_component(pconf, apr_psprintf(pconf, "mod_setenvifplus/%s", g_revision));
#ifdef SP_INTERNAL_TEST
  fprintf(stdout, "\033[1mmod_setenvifplus TEST BINARY, NOT FOR PRODUCTIVE USE\033[0m\n");
  fflush(stdout);
#endif
  return DECLINED;
}

static void sp_out(request_rec *r, apr_table_t *headers) {
  sp_config_t *dconf = ap_get_module_config(r->per_dir_config, &setenvifplus_module);
  sp_config_t *sconf = ap_get_module_config(r->server->module_config, &setenvifplus_module);

  if(sconf) {
    sp_header(r, headers, sconf->std_res_header);
  }
  if(dconf) {
    sp_setenv(r, headers, dconf->std_res_conditionals);
    sp_header(r, headers, dconf->std_res_header);
  }
  return;
}

static apr_status_t sp_out_err_filter(ap_filter_t *f, apr_bucket_brigade *bb) {
  request_rec *r = f->r;
  sp_out(r, r->err_headers_out);
  ap_remove_output_filter(f);
  return ap_pass_brigade(f->next, bb); 
}

static apr_status_t sp_out_filter(ap_filter_t *f, apr_bucket_brigade *bb) {
  request_rec *r = f->r;
  sp_out(r, r->headers_out);
  ap_remove_output_filter(f);
  return ap_pass_brigade(f->next, bb); 
}

/**
 * insert response filter
 */
static void sp_insert_filter(request_rec *r) {
  ap_add_output_filter("sp_out_filter", NULL, r, r->connection);
}

static void sp_insert_err_filter(request_rec *r) {
  ap_add_output_filter("sp_out_err_filter", NULL, r, r->connection);
}

/************************************************************************
 * directiv handlers 
 ***********************************************************************/
static void *sp_srv_config_create(apr_pool_t *p, server_rec *s) {
  sp_config_t *sconf = apr_pcalloc(p, sizeof(sp_config_t));
  sconf->path = NULL;
  sconf->std_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  sconf->std_late_conditionals = NULL;
  sconf->std_res_conditionals = NULL;
  sconf->std_req_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  sconf->std_late_req_header = NULL;
  sconf->std_res_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  sconf->user = apr_array_make(p, 2, sizeof(sp_usr_entry_t));
  return sconf;
}

static void *sp_srv_config_merge(apr_pool_t *p, void *basev, void *addv) {
  sp_config_t *sconf = apr_pcalloc(p, sizeof(sp_config_t));
  sp_config_t *base = basev;
  sp_config_t *add = addv;
  sconf->path = NULL;
  sconf->std_conditionals = apr_array_append(p, base->std_conditionals,
                                             add->std_conditionals);
  sconf->std_late_conditionals = NULL;
  sconf->std_res_conditionals = NULL;
  sconf->std_req_header = apr_array_append(p, base->std_req_header,
                                           add->std_req_header);
  sconf->std_late_req_header = NULL;
  sconf->std_res_header = apr_array_append(p, base->std_res_header,
                                           add->std_res_header);
  sconf->user = apr_array_append(p, base->user, add->user);
  return sconf;
}

static void *sp_dir_config_create(apr_pool_t *p, char *d) {
  sp_config_t *dconf = apr_pcalloc(p, sizeof(sp_config_t));
  dconf->path = d;
  dconf->std_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  dconf->std_late_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  dconf->std_res_conditionals = apr_array_make(p, 20, sizeof(sp_std_entry_t));
  dconf->std_req_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  dconf->std_late_req_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  dconf->std_res_header = apr_array_make(p, 20, sizeof(sp_hdr_entry_t));
  dconf->user = apr_array_make(p, 2, sizeof(sp_usr_entry_t));
  return dconf;
}

static void *sp_dir_config_merge(apr_pool_t *p, void *basev, void *addv) {
  sp_config_t *dconf = apr_pcalloc(p, sizeof(sp_config_t));
  sp_config_t *base = basev;
  sp_config_t *add = addv;
  dconf->path = add->path;
  dconf->std_conditionals = apr_array_append(p, base->std_conditionals,
                                             add->std_conditionals);
  dconf->std_late_conditionals = apr_array_append(p, base->std_late_conditionals,
                                                  add->std_late_conditionals);
  dconf->std_res_conditionals = apr_array_append(p, base->std_res_conditionals,
                                                 add->std_res_conditionals);
  dconf->std_req_header = apr_array_append(p, base->std_req_header,
                                           add->std_req_header);
  dconf->std_late_req_header = apr_array_append(p, base->std_late_req_header,
                                                add->std_late_req_header);
  dconf->std_res_header = apr_array_append(p, base->std_res_header,
                                           add->std_res_header);
  dconf->user = apr_array_append(p, base->user, add->user);
  return dconf;
}

const char *sp_header_cmd(cmd_parms *cmd, int argc, char *const argv[],
                          sp_hdr_entry_t *new) {
  const char *action;
  if(argc < 2) {
    /* always require action and header name */
    return apr_psprintf(cmd->pool, "%s: takes at least two arguments",
                        cmd->directive->directive);
  }
  action = argv[0];
  new->name = apr_pstrdup(cmd->pool, argv[1]);
  new->value = apr_pstrdup(cmd->pool, "");
  new->condition = NULL;
  if(strcasecmp(action, "set") == 0) {
    new->action = SP_ACTION_SET;
    if(argc < 3) {
      return apr_psprintf(cmd->pool, "%s set: takes at least three arguments",
                          cmd->directive->directive);
    }
    new->value = apr_pstrdup(cmd->pool, argv[2]);
    if(argc > 3) {
      char *c = argv[3];
      if(strncasecmp(c, "env=", strlen("env=")) == 0) {
        new->condition = apr_pstrdup(cmd->pool, &c[strlen("env=")]);
      } else {
        return apr_psprintf(cmd->pool, "%s set: expects 'env=' argument",
                            cmd->directive->directive);
      }
    }
  } else if(strcasecmp(action, "add") == 0) {
    new->action = SP_ACTION_ADD;
    if(argc < 3) {
      return apr_psprintf(cmd->pool, "%s add: takes at least three arguments",
                          cmd->directive->directive);
    }
    new->value = apr_pstrdup(cmd->pool, argv[2]);
    if(argc > 3) {
      char *c = argv[3];
      if(strncasecmp(c, "env=", strlen("env=")) == 0) {
        new->condition = apr_pstrdup(cmd->pool, &c[strlen("env=")]);
      } else {
        return apr_psprintf(cmd->pool, "%s add: expects 'env=' argument",
                            cmd->directive->directive);
      }
    }
  } else if(strcasecmp(action, "unset") == 0) {
    new->action = SP_ACTION_UNSET;
    if(argc > 2) {
      char *c = argv[2];
      if(strncasecmp(c, "env=", strlen("env=")) == 0) {
        new->condition = apr_pstrdup(cmd->pool, &c[strlen("env=")]);
      } else {
        return apr_psprintf(cmd->pool, "%s unset: expects 'env=' argument",
                            cmd->directive->directive);
      }
    }
  } else {
    return apr_psprintf(cmd->pool, "%s: unkown action",
                        cmd->directive->directive);
  }
  return NULL;
}

const char *sp_changeheader(cmd_parms *cmd, const char *header,
                            const char *regex, const char *value,
                            sp_hdr_entry_t *new) {
  new->action = SP_ACTION_CHANGE;
  new->preg = ap_pregcomp(cmd->pool, regex, (AP_REG_EXTENDED | AP_REG_ICASE));
  if(new->preg == NULL) {
    return apr_psprintf(cmd->pool, "%s: regex pattern '%s' could not be compied",
                        cmd->directive->directive, regex);
  }
  new->name = apr_pstrdup(cmd->pool, header);
  new->value = apr_pstrdup(cmd->pool, value);
  return NULL;
}

const char *sp_changereqheader_cmd(cmd_parms *cmd, void *dcfg, const char *header,
                                   const char *regex, const char *value) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->std_req_header);
  return sp_changeheader(cmd, header, regex, value, new);
}

const char *sp_changeresheader_cmd(cmd_parms *cmd, void *dcfg, const char *header,
                                   const char *regex, const char *value) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->std_res_header);
  return sp_changeheader(cmd, header, regex, value, new);
}

/**
 * RequestHeaderPlus 
 */
const char *sp_reqheader_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  if(strcmp(argv[argc-1], "late") == 0) {
    if(!cmd->path) {
      return apr_psprintf(cmd->pool, "%s: 'late' option not allowed here",
                          cmd->directive->directive);
    }
    new = apr_array_push(conf->std_late_req_header);
    argc--;
  } else {
    new = apr_array_push(conf->std_req_header);
  }
  return sp_header_cmd(cmd, argc, argv, new);
}

const char *sp_resheader_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  sp_hdr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->std_res_header);
  return sp_header_cmd(cmd, argc, argv, new);
}

const char *sp_std_core_cmd(cmd_parms *cmd, apr_array_header_t *conditionals,
                            const char *name, const char *regex, const char *var,
                            int icase, int late) {
  const char *val;
  const char *simple_pattern;
  sp_std_entry_t *new;

  /* create a new entry */
  new = apr_array_push(conditionals);
  new->name = name;
  new->regex = regex;
  new->icase = icase;
  if((simple_pattern = sp_non_regex_pattern(cmd->pool, regex))) {
    new->pattern = apr_strmatch_precompile(cmd->pool, simple_pattern, !icase);
    if(new->pattern == NULL) {
      return apr_psprintf(cmd->pool, "%s: strmatch pattern '%s' could not be compied",
                            cmd->directive->directive, simple_pattern);
    }
    new->preg = NULL;
  } else {
    new->preg = ap_pregcomp(cmd->pool, regex, (AP_REG_EXTENDED | (icase ? AP_REG_ICASE : 0)));
    if(new->preg == NULL) {
      return apr_psprintf(cmd->pool, "%s: regex pattern '%s' could not be compied",
                          cmd->directive->directive, regex);
    }
    new->pattern = NULL;
  }
  if(!strcasecmp(name, "remote_addr")) {
    new->special_type = SP_SPECIAL_REMOTE_ADDR;
  } else if(!strcasecmp(name, "remote_host")) {
    new->special_type = SP_SPECIAL_REMOTE_HOST;
  } else if(!strcasecmp(name, "request_uri")) {
    new->special_type = SP_SPECIAL_REQUEST_URI;
  } else if(!strcasecmp(name, "request_path")) {
    new->special_type = SP_SPECIAL_REQUEST_URI;
  } else if(!strcasecmp(name, "request_method")) {
    new->special_type = SP_SPECIAL_REQUEST_METHOD;
  } else if(!strcasecmp(name, "request_protocol")) {
    new->special_type = SP_SPECIAL_REQUEST_PROTOCOL;
  } else if(!strcasecmp(name, "server_addr")) {
    new->special_type = SP_SPECIAL_SERVER_ADDR;
  } else if(!strcasecmp(name, "request_query")) {
    new->special_type =  SP_SPECIAL_REQUEST_QUERY;
  } else if(!strcasecmp(name, "response_status")) {
    /* TODO: must be used for ResponseSetEnfIfPlus only */ 
    new->special_type =  SP_SPECIAL_RESPONSE_STATUS;
  } else if(!strcasecmp(name, "request_user")) {
    if(!late) {
      return apr_psprintf(cmd->pool, "%s: user match works only"
                          " when using the 'late' handler",
                          cmd->directive->directive);
    }
    new->special_type = SP_SPECIAL_REQUEST_USER;
  } else {
    new->special_type = SP_SPECIAL_NOT;
    /* Handle name as a regular expression.
     * If name a simple header string, identify as such
     * (new->pnamereg = NULL) to avoid the overhead of searching
     * through headers_in for a regex match.
     */
    if(sp_is_header_regex(cmd->pool, name)) {
      new->pnamereg = ap_pregcomp(cmd->pool, name,
                                  (AP_REG_EXTENDED | AP_REG_NOSUB
                                   | (icase ? AP_REG_ICASE : 0)));
      if (new->pnamereg == NULL)
      return apr_psprintf(cmd->pool, "%s: header name '%s' could not be compied",
                          cmd->directive->directive, name);
    } else {
      new->pnamereg = NULL;
    }
  }
  val = strchr(var, '=');
  if(val) {
    new->variable = apr_pstrndup(cmd->pool, var, val-var);
    val++;
    new->value = apr_pstrdup(cmd->pool, val);
  } else {
    new->variable = apr_pstrdup(cmd->pool, var);
    new->value = apr_pstrdup(cmd->pool, "1");
  }
  return NULL;
}

/**
 * SetEnvIfPlus
 * syntax: 
 * <attribute> <regex> [!]<env-variable>[=<value>] [late]
 */
const char *sp_std_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  int icase = cmd->info == SP_ICASE_MAGIC;
  sp_config_t *conf;
  apr_array_header_t *conditionals;
  int late = 0;
  const char *name;
  const char *regex;
  const char *var;
  if(argc < 3) {
    return apr_psprintf(cmd->pool, "%s: takes at least three arguments",
                        cmd->directive->directive);
  }
  name = argv[0];
  regex = argv[1];
  var = argv[2];
  /* determine scope of the directive */
  if(cmd->path) {
    if(argc > 4) {
      return apr_psprintf(cmd->pool, "%s: takes not more than four arguments",
                          cmd->directive->directive);
    }
    conf = dcfg;
    conditionals = conf->std_conditionals;
    if(argc == 4) {
      if(strcasecmp(argv[3], "late") == 0) {
        late = 1;
        conditionals = conf->std_late_conditionals;
      } else {
        return apr_psprintf(cmd->pool, "%s: unknown parameter '%s'",
                            cmd->directive->directive, argv[3]);
      }
    }
  } else {
    if(argc > 3) {
      return apr_psprintf(cmd->pool, "%s: takes not more than 3 arguments"
                          " when used at server level",
                          cmd->directive->directive);
    }
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
    conditionals = conf->std_conditionals;
  }

  return sp_std_core_cmd(cmd, conditionals, name, regex, var, icase, late);
}

/**
 * ResponseSetEnvIfPlus
 * syntax: 
 * <attribute> <regex> [!]<env-variable>[=<value>]
 */
const char *sp_res_std_cmd(cmd_parms *cmd, void *dcfg, int argc, char *const argv[]) {
  int icase = cmd->info == SP_ICASE_MAGIC;
  sp_config_t *conf = dcfg;
  int late = 0;
  const char *name;
  const char *regex;
  const char *var;
  if(argc != 3) {
    return apr_psprintf(cmd->pool, "%s: takes three arguments",
                        cmd->directive->directive);
  }
  name = argv[0];
  regex = argv[1];
  var = argv[2];
  return sp_std_core_cmd(cmd, conf->std_res_conditionals, name, regex, var, icase, late);
}

const char *sp_setuser_cmd(cmd_parms *cmd, void *dcfg, const char *header,
                           const char *regex, const char *value) {
  sp_usr_entry_t *new;
  sp_config_t *conf;
  if(cmd->path) {
    conf = dcfg;
  } else {
    conf = ap_get_module_config(cmd->server->module_config, &setenvifplus_module);
  }
  new = apr_array_push(conf->user);
  new->header = apr_pstrdup(cmd->pool, header);
  new->preg = ap_pregcomp(cmd->pool, regex, AP_REG_EXTENDED);
  if(new->preg == NULL) {
    return apr_psprintf(cmd->pool, "%s: regex pattern '%s' could not be compied",
                        cmd->directive->directive, regex);
  }
  new->value = apr_pstrdup(cmd->pool, value);
  return NULL;
}

static const command_rec sp_config_cmds[] = {
  AP_INIT_TAKE_ARGV("SetEnvIfPlus", sp_std_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "SetEnvIfPlus"
                    " <attribute> <regex> [!]<env-variable>[=<value>] [late],"
                    " defines an expression to apply to the specified request"
                    " attribute (request header) and sets the defined"
                    " environment variable of the expression matches."),
  AP_INIT_TAKE_ARGV("SetEnvIfPlusNoCase", sp_std_cmd, SP_ICASE_MAGIC,
                    RSRC_CONF|ACCESS_CONF,
                    "SetEnvIfPlusNoCase"
                    " <attribute> <regex> [!]<env-variable>[=<value>] [late]"
                    " defines an expression to apply to the specified request"
                    " attribute (request header) and sets the defined"
                    " environment variable of the expression matches."),
  AP_INIT_TAKE_ARGV("ResponseSetEnvIfPlus", sp_res_std_cmd, NULL,
                    ACCESS_CONF,
                    "ResponseSetEnvIfPlus"
                    " <attribute> <regex> [!]<env-variable>[=<value>],"
                    " defines an expression to apply to the specified response"
                    " attribute (response header) and sets the defined"
                    " environment variable of the expression matches."),
  AP_INIT_TAKE_ARGV("ResponseSetEnvIfPlusNoCase", sp_res_std_cmd, SP_ICASE_MAGIC,
                    ACCESS_CONF,
                    "ResponseSetEnvIfPlusNoCase"
                    " <attribute> <regex> [!]<env-variable>[=<value>]"
                    " defines an expression to apply to the specified response"
                    " attribute (response header) and sets the defined"
                    " environment variable of the expression matches."),
  AP_INIT_TAKE_ARGV("RequestHeaderPlus", sp_reqheader_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "RequestHeaderPlus set|unset|add <header> [<value>] [env=[!]<env-variable>] [late],"
                    " manipulates HTTP request headers. The first argument" 
                    " defines the action to take while the second argument"
                    " specifies the header name."
                    " The 'env=' attributes"
                    " is used to define additional conditions (variable must be set)."
                    " The 'late' attribute enforces late processing similar to"
                    " the SetEnvIfPlus directive."),
  AP_INIT_TAKE_ARGV("HeaderPlus", sp_resheader_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "HeaderPlus directive is deprecated (use RequestHeaderPlus)."),
  AP_INIT_TAKE_ARGV("ResponseHeaderPlus", sp_resheader_cmd, NULL,
                    RSRC_CONF|ACCESS_CONF,
                    "ResponseHeaderPlus set|unset|add <header> [<value>] [env=[!]<env-variable>],"
                    " manipulates HTTP request headers. The first argument" 
                    " defines the action to take while the second argument"
                    " specifies the header name."
                    " The 'env=' attributes"
                    " is used to define additional conditions (variable must be set)."),
  AP_INIT_TAKE3("ChangeRequestHeaderPlus", sp_changereqheader_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "ChangeRequestHeaderPlus <header> <regex> <value>, changes the"
                " value of the specified request header if the regex matches."),
  AP_INIT_TAKE3("ChangeResponseHeaderPlus", sp_changeresheader_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "ChangeResponseHeaderPlus <header> <regex> <value>, changes the"
                " value of the specified response header if the regex matches."),
  AP_INIT_TAKE3("SetUserPlus", sp_setuser_cmd, NULL,
                RSRC_CONF|ACCESS_CONF,
                "SetUserPlus <header> <regex> <value>,"
                " Used to set the an authenticated user name using the 'value'."),
  { NULL }
};

/************************************************************************
 * apache register 
 ***********************************************************************/
static void sp_register_hooks(apr_pool_t * p) {
  static const char *pre[] = { "mod_setenvif.c", "mod_ssl.c", NULL };
  static const char *post[] = { "mod_headers.c", NULL };
  ap_hook_post_read_request(sp_post_read_request, pre, post, APR_HOOK_MIDDLE);
  ap_hook_header_parser(sp_header_parser, pre, post, APR_HOOK_MIDDLE);
  ap_hook_check_user_id(sp_user_id, pre, NULL, APR_HOOK_FIRST);
  ap_hook_fixups(sp_fixup, pre, post, APR_HOOK_MIDDLE);
  ap_hook_post_config(sp_post_config, pre, NULL, APR_HOOK_MIDDLE);
#ifdef SP_INTERNAL_TEST
  ap_hook_handler(sp_handler, NULL, NULL, APR_HOOK_MIDDLE);
#endif
  ap_register_output_filter("sp_out_filter", sp_out_filter, NULL, AP_FTYPE_RESOURCE);
  ap_register_output_filter("sp_out_err_filter", sp_out_err_filter, NULL, AP_FTYPE_RESOURCE);
  ap_hook_insert_filter(sp_insert_filter, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_insert_error_filter(sp_insert_err_filter, NULL, NULL, APR_HOOK_MIDDLE);
}

/************************************************************************
 * apache module definition 
 ***********************************************************************/
module AP_MODULE_DECLARE_DATA setenvifplus_module ={ 
  STANDARD20_MODULE_STUFF,
  sp_dir_config_create,                    /**< dir config creater */
  sp_dir_config_merge,                     /**< dir merger */
  sp_srv_config_create,                    /**< server config */
  sp_srv_config_merge,                     /**< server merger */
  sp_config_cmds,                          /**< command table */
  sp_register_hooks,                       /**< hook registery */
};
